@@ -18,11 +18,10 @@ import subprocess
 import unittest
 from importlib import reload
 from logging import ERROR, WARNING
+from random import randint
 from typing import Optional
 from unittest import mock
 
-import pytest
-
 from opentelemetry import trace as trace_api
 from opentelemetry.context import Context
 from opentelemetry.sdk import resources, trace
@@ -37,12 +36,15 @@ from opentelemetry.sdk.trace import Resource, sampling
 from opentelemetry.sdk.trace.id_generator import RandomIdGenerator
 from opentelemetry.sdk.util import ns_to_iso_str
 from opentelemetry.sdk.util.instrumentation import InstrumentationInfo
+from opentelemetry.test.spantestutil import (
+    get_span_with_dropped_attributes_events_links,
+)
 from opentelemetry.trace import StatusCode
 from opentelemetry.util._time import _time_ns
 
 
-def new_tracer() -> trace_api.Tracer:
-    return trace.TracerProvider().get_tracer(__name__)
+def new_tracer(span_limits=None) -> trace_api.Tracer:
+    return trace.TracerProvider(span_limits=span_limits).get_tracer(__name__)
 
 
 class TestTracer(unittest.TestCase):
@@ -257,17 +259,25 @@ class TestSpanCreation(unittest.TestCase):
             tracer1 = tracer_provider.get_tracer("")
         with self.assertLogs(level=ERROR):
             tracer2 = tracer_provider.get_tracer(None)
-        self.assertEqual(
-            tracer1.instrumentation_info, tracer2.instrumentation_info
-        )
+
         self.assertIsInstance(
             tracer1.instrumentation_info, InstrumentationInfo
         )
         span1 = tracer1.start_span("foo")
         self.assertTrue(span1.is_recording())
         self.assertEqual(tracer1.instrumentation_info.version, "")
+        self.assertEqual(tracer1.instrumentation_info.name, "")
+
+        self.assertIsInstance(
+            tracer2.instrumentation_info, InstrumentationInfo
+        )
+        span2 = tracer2.start_span("bar")
+        self.assertTrue(span2.is_recording())
+        self.assertEqual(tracer2.instrumentation_info.version, "")
+        self.assertEqual(tracer2.instrumentation_info.name, "")
+
         self.assertEqual(
-            tracer1.instrumentation_info.name, "ERROR:MISSING MODULE NAME"
+            tracer1.instrumentation_info, tracer2.instrumentation_info
         )
 
     def test_span_processor_for_source(self):
@@ -530,6 +540,27 @@ class TestSpanCreation(unittest.TestCase):
             # pylint: disable=abstract-class-instantiated
             trace.Span("name", mock.Mock(spec=trace_api.SpanContext))
 
+    def test_surplus_span_links(self):
+        # pylint: disable=protected-access
+        max_links = trace.SpanLimits().max_links
+        links = [
+            trace_api.Link(trace_api.SpanContext(0x1, idx, is_remote=False))
+            for idx in range(0, 16 + max_links)
+        ]
+        tracer = new_tracer()
+        with tracer.start_as_current_span("span", links=links) as root:
+            self.assertEqual(len(root.links), max_links)
+
+    def test_surplus_span_attributes(self):
+        # pylint: disable=protected-access
+        max_attrs = trace.SpanLimits().max_attributes
+        attributes = {str(idx): idx for idx in range(0, 16 + max_attrs)}
+        tracer = new_tracer()
+        with tracer.start_as_current_span(
+            "span", attributes=attributes
+        ) as root:
+            self.assertEqual(len(root.attributes), max_attrs)
+
 
 class TestSpan(unittest.TestCase):
     # pylint: disable=too-many-public-methods
@@ -645,35 +676,6 @@ class TestSpan(unittest.TestCase):
                 isinstance(root.attributes["valid-byte-type-attribute"], str)
             )
 
-    def test_check_attribute_helper(self):
-        # pylint: disable=protected-access
-        self.assertFalse(trace._is_valid_attribute_value([1, 2, 3.4, "ss", 4]))
-        self.assertFalse(
-            trace._is_valid_attribute_value([dict(), 1, 2, 3.4, 4])
-        )
-        self.assertFalse(
-            trace._is_valid_attribute_value(["sw", "lf", 3.4, "ss"])
-        )
-        self.assertFalse(trace._is_valid_attribute_value([1, 2, 3.4, 5]))
-        self.assertTrue(trace._is_valid_attribute_value([1, 2, 3, 5]))
-        self.assertTrue(trace._is_valid_attribute_value([1.2, 2.3, 3.4, 4.5]))
-        self.assertTrue(trace._is_valid_attribute_value([True, False]))
-        self.assertTrue(trace._is_valid_attribute_value(["ss", "dw", "fw"]))
-        self.assertTrue(trace._is_valid_attribute_value([]))
-        self.assertFalse(trace._is_valid_attribute_value(dict()))
-        self.assertTrue(trace._is_valid_attribute_value(True))
-        self.assertTrue(trace._is_valid_attribute_value("hi"))
-        self.assertTrue(trace._is_valid_attribute_value(3.4))
-        self.assertTrue(trace._is_valid_attribute_value(15))
-        # None in sequences are valid
-        self.assertTrue(trace._is_valid_attribute_value(["A", None, None]))
-        self.assertTrue(
-            trace._is_valid_attribute_value(["A", None, None, "B"])
-        )
-        self.assertTrue(trace._is_valid_attribute_value([None, None]))
-        self.assertFalse(trace._is_valid_attribute_value(["A", None, 1]))
-        self.assertFalse(trace._is_valid_attribute_value([None, "A", None, 1]))
-
     def test_sampling_attributes(self):
         sampling_attributes = {
             "sampler-attr": "sample-val",
@@ -806,7 +808,7 @@ class TestSpan(unittest.TestCase):
             self.assertEqual(
                 root.links[0].context.span_id, other_context1.span_id
             )
-            self.assertEqual(root.links[0].attributes, None)
+            self.assertEqual(0, len(root.links[0].attributes))
             self.assertEqual(
                 root.links[1].context.trace_id, other_context2.trace_id
             )
@@ -815,6 +817,9 @@ class TestSpan(unittest.TestCase):
             )
             self.assertEqual(root.links[1].attributes, {"name": "neighbor"})
 
+            with self.assertRaises(TypeError):
+                root.links[1].attributes["name"] = "new_neighbour"
+
     def test_update_name(self):
         with self.tracer.start_as_current_span("root") as root:
             # name
@@ -870,7 +875,7 @@ class TestSpan(unittest.TestCase):
         self.assertEqual(end_time, span.end_time)
 
     def test_ended_span(self):
-        """"Events, attributes are not allowed after span is ended"""
+        """Events, attributes are not allowed after span is ended"""
 
         root = self.tracer.start_span("root")
 
@@ -914,7 +919,6 @@ class TestSpan(unittest.TestCase):
             with self.assertRaises(AssertionError):
                 with context as root:
                     raise AssertionError("unknown")
-
             self.assertIs(root.status.status_code, StatusCode.ERROR)
             self.assertEqual(
                 root.status.description, "AssertionError: unknown"
@@ -929,18 +933,53 @@ class TestSpan(unittest.TestCase):
             .start_as_current_span("root")
         )
 
-    def test_last_status_wins(self):
+    def test_status_cannot_override_ok(self):
         def error_status_test(context):
             with self.assertRaises(AssertionError):
                 with context as root:
                     root.set_status(trace_api.status.Status(StatusCode.OK))
                     raise AssertionError("unknown")
+            self.assertIs(root.status.status_code, StatusCode.OK)
+            self.assertIsNone(root.status.description)
 
+        error_status_test(
+            trace.TracerProvider().get_tracer(__name__).start_span("root")
+        )
+        error_status_test(
+            trace.TracerProvider()
+            .get_tracer(__name__)
+            .start_as_current_span("root")
+        )
+
+    def test_status_cannot_set_unset(self):
+        def unset_status_test(context):
+            with self.assertRaises(AssertionError):
+                with context as root:
+                    raise AssertionError("unknown")
+            root.set_status(trace_api.status.Status(StatusCode.UNSET))
             self.assertIs(root.status.status_code, StatusCode.ERROR)
             self.assertEqual(
                 root.status.description, "AssertionError: unknown"
             )
 
+        unset_status_test(
+            trace.TracerProvider().get_tracer(__name__).start_span("root")
+        )
+        unset_status_test(
+            trace.TracerProvider()
+            .get_tracer(__name__)
+            .start_as_current_span("root")
+        )
+
+    def test_last_status_wins(self):
+        def error_status_test(context):
+            with self.assertRaises(AssertionError):
+                with context as root:
+                    raise AssertionError("unknown")
+                root.set_status(trace_api.status.Status(StatusCode.OK))
+                self.assertIs(root.status.status_code, StatusCode.OK)
+                self.assertIsNone(root.status.description)
+
         error_status_test(
             trace.TracerProvider().get_tracer(__name__).start_span("root")
         )
@@ -1221,7 +1260,10 @@ class TestSpanProcessor(unittest.TestCase):
             is_remote=False,
             trace_flags=trace_api.TraceFlags(trace_api.TraceFlags.SAMPLED),
         )
-        span = trace._Span("span-name", context, resource=Resource({}))
+        parent = trace._Span("parent-name", context, resource=Resource({}))
+        span = trace._Span(
+            "span-name", context, resource=Resource({}), parent=parent
+        )
 
         self.assertEqual(
             span.to_json(),
@@ -1233,7 +1275,7 @@ class TestSpanProcessor(unittest.TestCase):
         "trace_state": "[]"
     },
     "kind": "SpanKind.INTERNAL",
-    "parent_id": null,
+    "parent_id": "0x00000000deadbef0",
     "start_time": null,
     "end_time": null,
     "status": {
@@ -1247,7 +1289,7 @@ class TestSpanProcessor(unittest.TestCase):
         )
         self.assertEqual(
             span.to_json(indent=None),
-            '{"name": "span-name", "context": {"trace_id": "0x000000000000000000000000deadbeef", "span_id": "0x00000000deadbef0", "trace_state": "[]"}, "kind": "SpanKind.INTERNAL", "parent_id": null, "start_time": null, "end_time": null, "status": {"status_code": "UNSET"}, "attributes": {}, "events": [], "links": [], "resource": {}}',
+            '{"name": "span-name", "context": {"trace_id": "0x000000000000000000000000deadbeef", "span_id": "0x00000000deadbef0", "trace_state": "[]"}, "kind": "SpanKind.INTERNAL", "parent_id": "0x00000000deadbef0", "start_time": null, "end_time": null, "status": {"status_code": "UNSET"}, "attributes": {}, "events": [], "links": [], "resource": {}}',
         )
 
     def test_attributes_to_json(self):
@@ -1270,17 +1312,56 @@ class TestSpanProcessor(unittest.TestCase):
 
 
 class TestSpanLimits(unittest.TestCase):
-    @mock.patch.dict(
-        "os.environ",
-        {
-            OTEL_SPAN_ATTRIBUTE_COUNT_LIMIT: "10",
-            OTEL_SPAN_EVENT_COUNT_LIMIT: "20",
-            OTEL_SPAN_LINK_COUNT_LIMIT: "30",
-        },
-    )
-    def test_span_environment_limits(self):
-        reload(trace)
-        tracer = new_tracer()
+    # pylint: disable=protected-access
+
+    def test_limits_defaults(self):
+        limits = trace.SpanLimits()
+        self.assertEqual(
+            limits.max_attributes,
+            trace._DEFAULT_OTEL_SPAN_ATTRIBUTE_COUNT_LIMIT,
+        )
+        self.assertEqual(
+            limits.max_events, trace._DEFAULT_OTEL_SPAN_EVENT_COUNT_LIMIT
+        )
+        self.assertEqual(
+            limits.max_links, trace._DEFAULT_OTEL_SPAN_LINK_COUNT_LIMIT
+        )
+
+    def test_limits_values_code(self):
+        max_attributes, max_events, max_links = (
+            randint(0, 10000),
+            randint(0, 10000),
+            randint(0, 10000),
+        )
+        limits = trace.SpanLimits(
+            max_attributes=max_attributes,
+            max_events=max_events,
+            max_links=max_links,
+        )
+        self.assertEqual(limits.max_attributes, max_attributes)
+        self.assertEqual(limits.max_events, max_events)
+        self.assertEqual(limits.max_links, max_links)
+
+    def test_limits_values_env(self):
+        max_attributes, max_events, max_links = (
+            randint(0, 10000),
+            randint(0, 10000),
+            randint(0, 10000),
+        )
+        with mock.patch.dict(
+            "os.environ",
+            {
+                OTEL_SPAN_ATTRIBUTE_COUNT_LIMIT: str(max_attributes),
+                OTEL_SPAN_EVENT_COUNT_LIMIT: str(max_events),
+                OTEL_SPAN_LINK_COUNT_LIMIT: str(max_links),
+            },
+        ):
+            limits = trace.SpanLimits()
+            self.assertEqual(limits.max_attributes, max_attributes)
+            self.assertEqual(limits.max_events, max_events)
+            self.assertEqual(limits.max_links, max_links)
+
+    def _test_span_limits(self, tracer):
         id_generator = RandomIdGenerator()
         some_links = [
             trace_api.Link(
@@ -1292,14 +1373,123 @@ class TestSpanLimits(unittest.TestCase):
             )
             for _ in range(100)
         ]
-        with pytest.raises(ValueError):
-            with tracer.start_as_current_span("root", links=some_links):
-                pass
 
-        with tracer.start_as_current_span("root") as root:
+        some_attrs = {
+            "init_attribute_{}".format(idx): idx for idx in range(100)
+        }
+        with tracer.start_as_current_span(
+            "root", links=some_links, attributes=some_attrs
+        ) as root:
+            self.assertEqual(len(root.links), 30)
+            self.assertEqual(len(root.attributes), 10)
             for idx in range(100):
                 root.set_attribute("my_attribute_{}".format(idx), 0)
                 root.add_event("my_event_{}".format(idx))
 
             self.assertEqual(len(root.attributes), 10)
             self.assertEqual(len(root.events), 20)
+
+    def _test_span_no_limits(self, tracer):
+        num_links = int(trace._DEFAULT_OTEL_SPAN_LINK_COUNT_LIMIT) + randint(
+            1, 100
+        )
+
+        id_generator = RandomIdGenerator()
+        some_links = [
+            trace_api.Link(
+                trace_api.SpanContext(
+                    trace_id=id_generator.generate_trace_id(),
+                    span_id=id_generator.generate_span_id(),
+                    is_remote=False,
+                )
+            )
+            for _ in range(num_links)
+        ]
+        with tracer.start_as_current_span("root", links=some_links) as root:
+            self.assertEqual(len(root.links), num_links)
+
+        num_events = int(trace._DEFAULT_OTEL_SPAN_EVENT_COUNT_LIMIT) + randint(
+            1, 100
+        )
+        with tracer.start_as_current_span("root") as root:
+            for idx in range(num_events):
+                root.add_event("my_event_{}".format(idx))
+
+            self.assertEqual(len(root.events), num_events)
+
+        num_attributes = int(
+            trace._DEFAULT_OTEL_SPAN_ATTRIBUTE_COUNT_LIMIT
+        ) + randint(1, 100)
+        with tracer.start_as_current_span("root") as root:
+            for idx in range(num_attributes):
+                root.set_attribute("my_attribute_{}".format(idx), 0)
+
+            self.assertEqual(len(root.attributes), num_attributes)
+
+    @mock.patch.dict(
+        "os.environ",
+        {
+            OTEL_SPAN_ATTRIBUTE_COUNT_LIMIT: "10",
+            OTEL_SPAN_EVENT_COUNT_LIMIT: "20",
+            OTEL_SPAN_LINK_COUNT_LIMIT: "30",
+        },
+    )
+    def test_span_limits_env(self):
+        self._test_span_limits(new_tracer())
+
+    @mock.patch.dict(
+        "os.environ",
+        {
+            OTEL_SPAN_ATTRIBUTE_COUNT_LIMIT: "10",
+            OTEL_SPAN_EVENT_COUNT_LIMIT: "20",
+            OTEL_SPAN_LINK_COUNT_LIMIT: "30",
+        },
+    )
+    def test_span_limits_default_to_env(self):
+        self._test_span_limits(
+            new_tracer(
+                span_limits=trace.SpanLimits(
+                    max_attributes=None, max_events=None, max_links=None
+                )
+            )
+        )
+
+    def test_span_limits_code(self):
+        self._test_span_limits(
+            new_tracer(
+                span_limits=trace.SpanLimits(
+                    max_attributes=10, max_events=20, max_links=30
+                )
+            )
+        )
+
+    @mock.patch.dict(
+        "os.environ",
+        {
+            OTEL_SPAN_ATTRIBUTE_COUNT_LIMIT: "unset",
+            OTEL_SPAN_EVENT_COUNT_LIMIT: "unset",
+            OTEL_SPAN_LINK_COUNT_LIMIT: "unset",
+        },
+    )
+    def test_span_no_limits_env(self):
+        self._test_span_no_limits(new_tracer())
+
+    def test_span_no_limits_code(self):
+        self._test_span_no_limits(
+            new_tracer(
+                span_limits=trace.SpanLimits(
+                    max_attributes=trace.SpanLimits.UNSET,
+                    max_links=trace.SpanLimits.UNSET,
+                    max_events=trace.SpanLimits.UNSET,
+                )
+            )
+        )
+
+    def test_dropped_attributes(self):
+        span = get_span_with_dropped_attributes_events_links()
+        self.assertEqual(1, span.dropped_links)
+        self.assertEqual(2, span.dropped_attributes)
+        self.assertEqual(3, span.dropped_events)
+        self.assertEqual(2, span.events[0].attributes.dropped)
+        self.assertEqual(2, span.links[0].attributes.dropped)
+        self.assertEqual(2, span.resource.attributes.dropped)
